#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class InventoryDevice < ApplicationRecord
  include Redborder::Audit

  belongs_to :sensor
  belongs_to :inventory_device_type_object, optional: true
  belongs_to :operating_system_object, optional: true
  belongs_to :hardware_object, optional: true
  belongs_to :criticality, optional: true

  has_many :mac_objects, dependent: :destroy

  has_and_belongs_to_many :incidents
  
  before_create :set_default_criticality

  after_create :assign_sensor_to_macs
  after_update :assign_sensor_to_macs

  accepts_nested_attributes_for :mac_objects, allow_destroy: true

  validate :ips_must_be_valid

  audited on: [:create, :update, :destroy]

  # ================================================================
  # Public cache methods
  # ================================================================
  def self.cache_key_for_user(user_id)
    "rb-webui:inventory-devices:#{user_id}"
  end
  
  # Save the current filters to Memcached
  def self.create_filters(user_id, search_params, name_param)
    filters = { search: search_params, name: name_param }
    Rails.cache.write(cache_key_for_user(user_id), filters, expires_in: 3.hours)
  end
  
  # Read the filters from Memcached
  def self.read_filters(user_id)
    Rails.cache.read(cache_key_for_user(user_id))
  end

  # Clear the filters from Memcached
   def self.clear_filters(user_id)
    Rails.cache.delete(cache_key_for_user(user_id))
  end

  def criticality_name
    self.criticality ? self.criticality.name : 'unset'
  end

  def to_asset(mac)
    type = self.inventory_device_type_object.name rescue nil
    os = self.operating_system_object.short_name rescue nil
    {
      id: self.id,
      name: self.name,
      mac_address: mac,
      type: type,
      operating_system: os
    }
  end

  def ips_must_be_valid
    return if ips.blank?

    ip_example = "(#{I18n.t('example_short')} #{I18n.t('ip_v4_example')} / #{I18n.t('ip_v6_example')} )"
  
    ips.each do |ip_value|
      unless ip_value.match?(Regexp.new(ip_pattern))
        errors.add(:ips, "#{ip_value} #{I18n.t('errors.messages.invalid')}\n#{ip_example}")
        next
      end
  
      begin
        IPAddr.new(ip_value)
      rescue IPAddr::InvalidAddressError, ArgumentError
        errors.add(:ips, "#{ip_value} #{I18n.t('errors.messages.invalid')}\n#{ip_example}")
      end
    end
  end

  def ip_pattern
    "^((([0-9]{1,3}\\.){3}[0-9]{1,3})|(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)))$"
  end

  # Change the sensors of the inventory device's mac objects
  def assign_sensor_to_macs
    mac_objects.each do |mac|
      mac.sensor_id = self.sensor_id
      mac.save!
    end
  end

  # Change the inventory device sensor and all of the sensor of its mac objects
  def propagate_sensor new_sensor
    # This is only required when a mac change its sensor and executes this method
    # to change the asset sensor.
    if self.sensor_id != new_sensor
      self.sensor_id = new_sensor
      save!
    end
    assign_sensor_to_macs
  end
  
  def create_incident_and_notify_vault
    return unless self.inventory_device_type_object && self.inventory_device_type_object.auto_create_incident
    
    Delayed::Worker.logger.info "Creating incident.."
    # Create incident
    incident = Incident.new
    incident.name = "New #{self.inventory_device_type_object.name} detected"
    incident.name += " id: #{self.id}" if self.id.present?
    incident.name += " name: #{self.name}" if self.name.present?
    incident.name += " macs: #{self.mac_objects.map(&:value).join(", ")}" if self.mac_objects.any?
    incident.name += " ips: #{self.ips.join(', ')}" if self.ips.any?
    incident.priority = map_criticality_to_vault_priority(self.inventory_device_type_object.criticality.name.underscore)
    incident.uuid = SecureRandom.uuid
    incident.incident_status = IncidentStatus.find_by(id: 4)  # 4 = NEW
    if Redborder::Module.registered_modules[:vault] && Redborder::Module.registered_modules[:vault][:incident_source_name]
      incident.source = Redborder::Module.registered_modules[:vault][:incident_source_name]
    end
    incident.sensor_id = self.sensor.id
    if incident.save
      begin
        Delayed::Worker.logger.info "Incident created with id: #{incident.id} and name: #{incident.name}"
        syslog_message = Rfc5424SyslogMessage.new
        syslog_message.app_name = 'inventorydevicesjobs'
        syslog_message.msg_id = 'new-inventory-device'
        syslog_message.structured_data.push(new_inventory_device_event_structured_data(incident))
        # TODO: let the user choose the severity
        syslog_message.severity = 4 # 4 = Warning
        syslog_message.msg = incident.name

        params_syslog = { syslog_server: '127.0.0.1' }
        Delayed::Worker.logger.info "Sending Syslog alert: #{syslog_message.to_s}"
        SyslogService.send_to_syslog(syslog_message.to_s, params_syslog)

      rescue => e
        Rails.logger.error "Error while sending Syslog alert: #{e.message}"
      end
    end
  rescue => e
    Rails.logger.error "Error in notify_vault: #{e.message} - Backtrace: #{e.backtrace.join("\n")}"
  end

  def new_inventory_device_event_structured_data(incident)
    sd_id = "new-inventory-device-event@#{RedBorder::PEN}"

    fields = {
      "incident-uuid" => incident.uuid,
      "incident-uri" => incident.uri,
      "incident-domain" => incident.domain,
      "sensor-name" => self.sensor.name,
      "sensor-uuid" => self.sensor.uuid
    }
   
    service_provider = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::SERVICE_PROVIDER)
    deployment       = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::DEPLOYMENT)
    namespace        = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::NAMESPACE)
    organization     = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::ORGANIZATION)
    market           = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::MARKET)
    campus           = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::CAMPUS)
    building         = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::BUILDING)
    floor            = self.sensor.self_and_ancestors.find_by(domain_type: Sensor::FLOOR)

    %w(service_provider deployment namespace organization market campus building floor).each do |v|
      fields["#{v}"] = eval(v).try('name')
      fields["#{v}-uuid"] = eval(v).try('uuid')
    end

    "[#{sd_id} " + fields.map { |key, value| "#{key}=\"#{value}\"" }.join(" ") + "]"
  end

  private

  def set_default_criticality
    self.criticality ||= Criticality.find_or_create_by(name: "Unset")
  end

  def map_criticality_to_vault_priority(criticality_name)
    case criticality_name
    when 'unset'
      return Priority.find_by_name('debug')
    when 'none'
      return Priority.find_by_name('info')
    when 'low'
      return Priority.find_by_name('notice')
    when 'medium'
      return Priority.find_by_name('warning')
    when 'high'
      return Priority.find_by_name('error')
    when 'critical'
      return Priority.find_by_name('critical')
    else
      return Priority.find_by_name('unknown')
    end
  end
end